(**
   This is a main module of CO3.

@author Naoki Nishida
@since 2014/06/26
@since 2016/08/28
@since 2016/11/04
@since 2017/01/06
@since 2017/08/01
@since 2018/06/12
*)

open Debug;;
open Ctrs;;

(**
   Type for execution mode.
*)
type exec = 
    SimU 
  | SeqU 
  | SRn 
  | SR 
  | CO2 
  | CO3 
  | CO3verbose 
  | SNofTRS
  | CheckProperties;;

(**
  Exception for other processes.
*)
exception Others;;

(**
   Mode for execution.
*)
let mode = ref CO3;;
    
(**
   a function to prove termination.
*)
let prove_SN = ref Sn.is_SN;;

(**
   Flag for competition mode.
*)
let competition = ref false;;
    
(**
   This sets a mode via the name of the executed command.
*)
let set_mode com =
  match com with
    "./co2un" -> mode := SimU
  | "./co2u" -> mode := SeqU
  | "./co2srn" -> mode := SRn
  | "./co2srnc" ->
      begin
	Convert_NormalOneCTRS.set_omitting_brackets ();
	mode := SRn
      end
  | "./co2sr" -> mode := SR
  | "./co2" -> mode := CO2
  | "./co3" -> mode := CO3
  | "./co3v" -> mode := CO3verbose
  | "./co3d" -> 
      begin
	set_debugmode ();
	mode := CO3
      end
  | "./co3coco" -> 
      begin
	competition := true;
(*
	set_debugmode ();
*)
	Convert_NormalOneCTRS.set_omitting_brackets ();
	mode := CO3
      end
  | "./co3sn" -> mode := SNofTRS
  | "./co3natt" ->
      begin
	prove_SN := Sn.is_SN_by_natt;
	mode := SNofTRS
      end
  | "./co3aprove" ->
      begin
	prove_SN := Sn.is_SN_by_aprove;
	mode := SNofTRS
      end
  | "./co3chk" -> mode := CheckProperties
  | _ -> raise Others;;
	
(**
   This checks whether [mode] is for CO3.
*)
let is_co3 () =
  match !mode with
    CO3
  | CO3verbose
    -> true
  | _ -> false;;

(**
   This is an exception for commands to finish the process.
*)
exception Finish;;

(**
   This is an exception for failing to prove confluence.
*)
exception Fail_toProve_Confluence;;

let _ =
  let output = ref "" in
  let result_Un_sound = ref false in
  let result_SRn_sound = ref false in
  let result_U_sound = ref false in
  let result_SR_sound = ref false in
  let append_output s =  output := (!output)^"\n"^s  in
  let debug s = begin Debug.debug s; append_output s end in
  let print_output () =
    let localprint =
        if !competition then prerr_string else print_string
    in
    if !mode = CO3verbose || !mode = CO2 || !competition
    then
      begin
	localprint "\n";
	if !mode = CO3verbose
	then
	  localprint "------------------------------------------------------------\n"
	else
	  ();
	localprint (!output);
	if !mode = CO3verbose
	then
	  localprint "------------------------------------------------------------\n"
	else
	  ();
	()
      end
    else
      ()
  in
  let check_args () =
     (* This checks what command-line arguments are. *)
    if Array.length Sys.argv <= 1 || Array.length Sys.argv > 3
    then
      false
    else
      true
  in
  try
    (* Check out the number of command-line arguments *)
    if check_args () (* Array.length Sys.argv = 2 *)
    then
      begin
        (* Set a mode from the command name. *)
	set_mode Sys.argv.(0);
        (* Read .trs file *)
	let result = 
	  try
	    Readtrs.readtrs (Sys.argv.(1))
	  with
	    _ -> raise Unsupported
	in
	debug ("Succeeded in reading \""^Sys.argv.(1)^"\".\n"^(string_of_ctrs 2 result));
        (* Check whether an ->= rule is included or non-standard *)
	if has_relrew result || not (is_standard result)
	then 
	  raise Unsupported 
	else ();
	debug "No \"->=\"-rules.\n";
        if !mode = CheckProperties
        then
          begin
            if Check_properties.check result
            then
              print_string "YES\n"
            else
              print_string "NO\n";
            raise Finish
          end
        else
          ();
        (* This constraint is for adapting this to CoCo.
           raise Unsupported if given R is a TRS or not "oriented 3-CTRS" *)
	if (!competition
	      && ((is_unconditional result 
		 || not (is_type_three result && result.ctype = (Some Oriented)))))
	then
	  begin
	    match (is_type_three result, result.ctype = (Some Oriented)) with
	      (true,true) -> (* Not happend *)
		begin
		  debug "Type 3 and oriented\n";
		  raise Unsupported
		end
	    | (true,false) ->
		begin
		  debug "Type 3 but not oriented\n";
		  raise Unsupported
		end
	    | (false,true) ->
		begin
		  debug "Not type 3 but oriented\n";
		  raise Unsupported
		end
	    | (false,false) ->
		begin
		  debug "Neither type 3 nor oriented\n";
		  raise Unsupported
		end
	  end
	else 
	  ();

	if !mode = SNofTRS
	then
	  if is_unconditional result 
	  then
	    begin
	      debug ("Try to prove termination of the following TRS:\n"^(string_of_ctrs 2 result));
	      begin
		match !prove_SN result with
		  Some true ->
		    begin
		      debug "Suceeded in proving termination.\n";
		      print_string "YES\n";
		      raise Finish
		    end
		| Some false ->
		    begin
		      debug "Suceeded in disproving termination.\n";
		      print_string "NO\n";
		      raise Finish
		    end
		| None ->
		    begin
		      debug "Failed to prove termination.\n";
		      print_string "MAYBE\n";
		      raise Finish
		    end
	      end
	    end
	  else
	    begin
	      debug "The input is not a TRS.\n";
	      raise Unsupported
	    end
	else
	  ();

        (* For proving confluence.
	   Decompose conditions: c(s1,...,sn) == c(t1,...,tn) to s1 == t1, ..., sn == tn *)
	let result = 
	  if is_co3 ()
	  then
	    let result0 = decompose_conditions result in
	    begin
	      debug ("Decomposed conditions if possible.\n"^(string_of_ctrs 2 result0));
	      result0
	    end
	  else
	    result
	in
        (* For proving confluence.
	   Remove infeasible rules *)
	let result =
	  if is_co3 ()
	  then
	    let result0 = to_trs (Infeasibility.rm_infeasible_rules_amap result) in
	    begin
	      debug ("Removed infeasible rules as much as possible.\n"^(string_of_ctrs 2 result0));
	      result0
	    end
	  else
	    result
	in
        (* For proving confluence.
	   Check first non-confluence *)
	if is_co3 () 
	    && 
	  begin
	    debug ("Try to disprove confluence of the following (C)TRS:\n"^(string_of_ctrs 2 result));
	    Cr.is_not_CR result
	  end
	then 
	  begin
	    debug "Succeeded in disproving confluence.\n";
	    append_output "Disproved via the following (C)TRS:";
	    append_output (Ctrs.string_of_ctrs 0 result);
	    append_output "NO\n";
	    print_string "NO\n";
	    print_output ()
	  end
	else (* Case where failed to prove non-confluence of R *)
	  (* Check whether R is a TRS or not *)
	  if is_unconditional result
	  then 
	    begin
	      debug ("Try to prove confluence of the following TRS:\n"^(string_of_ctrs 2 result));
              (* Improved for CoCo 2015 *)
	      (* Check whether R is orthogonal *)
	      if Cr.is_orthogonal result 
	      then 
		begin
		  debug "Succeeded in proving confluence due to orthogonality.\n";
		  append_output "Proved via the following orthogonal TRS:";
		  append_output (Ctrs.string_of_ctrs 0 result);
		  append_output "YES\n";
		  print_string "YES\n";
		  print_output ()
		end
	      else
		begin
		  debug "Not othogonal.\n";
		  debug "Try to prove termination.\n";
	          (* Check whether R is SN and CP-joinable *)
		  if Sn.is_SN result = Some true
		  then
		    begin
		      debug "Succeeded in proving termination.\n";
		      debug "Try to prove joinability of CPs.\n";
		      if List.for_all (Cpair.is_joinable result) (Cpair.all_cpairs result)
		      then
			begin
			  debug "Succeeded in proving joinablity of all CPs.\n";
			  append_output "Proved via the following terminating and CP-joinable TRS:";
			  append_output (Ctrs.string_of_ctrs 0 result);
			  append_output "YES\n";
			  print_string "YES\n";
			  print_output ()
			end
		      else 
			begin
			  debug "Exists a CP that is not joinable.\n";
			  append_output "Disproved via the following TRS:";
			  append_output (Ctrs.string_of_ctrs 0 result);
			  append_output "NO\n";
			  print_string "NO\n";
			  print_output ()
			end
		    end
		  else
		    begin
		      debug "Failed to prove termination.\n";
		      append_output "Failed via the following TRS:";
		      append_output (Ctrs.string_of_ctrs 0 result);
		      append_output "MAYBE\n";
		      print_string "MAYBE\n";
		      print_output ()
		    end
		end
	    end
	  else (* Case where R is a CTRS *)
	    begin
	      try
                raise Fail_toProve_Confluence;
	        (* Check whether SR is sound for R *)
		if !mode = SimU || !mode = SRn
		then
		  begin
		    if Convert_NormalOneCTRS.is_SR_sound result
		    then
		      begin
			debug "SR for normal 1CTRSs is sound for this CTRS.\n";
			result_SRn_sound := true 
		      end
		    else 
			debug "SR for normal 1CTRSs is not sound for this CTRS.\n";
		    if Convert_NormalOneCTRS.is_U_sound result
		    then 
		      begin
			debug "The simulateneous U is sound for this CTRS.\n";
			result_Un_sound := true
		      end
		    else
		      debug "The simulateneous U is not sound for this CTRS.\n";
		    ()
		  end
		else
		  ();
     	        (* Convert R to SR(R) and U(R) *)
		debug "Try to apply SR and U for normal 1CTRSs to the above CTRS.\n";
		let (result2,result3) = Convert_NormalOneCTRS.convert result in
		debug ("Succeeded in applying SR and U for normal 1CTRSs to the above CTRS.\n"
		       ^"SR_n(R) =\n"^(Ctrs.string_of_ctrs 2 result2)
		       ^"U_n(R) =\n"^(Ctrs.string_of_ctrs 2 result3));
		if !mode = SRn
		then
		  begin
		    print_string (Ctrs.string_of_ctrs 0 result2);
		    raise Finish
		  end
		else
		  ();
		if !mode = SimU
		then
		  begin
		    print_string (Ctrs.string_of_ctrs 0 result3);
		    raise Finish
		  end
		else
		  ();
		if !mode = CO3verbose || !mode = CO2
		then
		  begin
		    append_output "SR_n(R) =";
		    append_output (Ctrs.string_of_ctrs 0 result2);
		    append_output "\nU_n(R) =";
		    append_output (Ctrs.string_of_ctrs 0 result3);
		    ()
		  end
		else
		  ();
		if is_co3 () 
		then
		  if not !competition
                        &&
                     !result_Un_sound 
			&&
		      begin
			debug "The simulataneous U is sound for the above CTRS.\n";
			Cr.is_CR result3
		      end
		  then
		    begin
		      debug "U_n(R) is confluent.\n";
		      append_output "YES\n";
		      print_string "YES\n";
		      print_output ()
		    end
                  else
		    if !result_SRn_sound 
		        &&
		      begin
		        debug "SR for normal 1CTRSs is sound for the above CTRS.\n";
		        Cr.is_CR result2
		      end
		    then
		      begin
		        debug "SR_n(R) is confluent.\n";
		        append_output "YES\n";
		        print_string "YES\n";
		        print_output ()
		      end
		    else
		      raise Fail_toProve_Confluence
		else
		  raise Not_normal_CTRS (* Not necessary? *)
	      with
		Fail_toProve_Confluence
	      |	Not_normal_CTRS ->
		  debug "Failed either to apply SR and U for normal 1CTRSs to the above CTRS or to prove confluence of any converted TRSs.\n";
	          (* Check whether SR is sound for R that is not a normal 1-CTRS  *)
		  if Convert_UWLLDCTRS.is_SR_sound result 
		  then
		    result_SR_sound := true
		  else
		    ();
		  if Convert_UWLLDCTRS.is_U_sound result
		  then
		    result_U_sound := true
		  else
		    ();
	          (* Convert R to SR(R) and U(R) *)
		  debug "Try to apply SR and U for 3DCTRSs to the above CTRS.\n";
		  let (result2,result3) = Convert_UWLLDCTRS.convert result in
(*		  debug ("Succeeded in applying SR and U for 3DCTRSs to the above CTRS.\n" *)
		  debug ("Succeeded in applying U for 3DCTRSs to the above CTRS.\n"
(*		       ^"SR(R) =\n"^(Ctrs.string_of_ctrs 2 result2) *)
		       ^"U(R) =\n"^(Ctrs.string_of_ctrs 2 result3));
		  if !mode = SR
		  then
		    begin
		      print_string (Ctrs.string_of_ctrs 0 result2);
		      raise Finish
		    end
		  else
		    ();
		  if !mode = SeqU
		  then
		    begin
		      print_string (Ctrs.string_of_ctrs 0 result3);
		      raise Finish
		    end
		  else
		    ();
		  if !mode = CO3verbose || !mode = CO2
		  then
		    begin
                      append_output "SR(R) =";
		      append_output (Ctrs.string_of_ctrs 0 result2);
                      append_output "\nU(R) =";
		      append_output (Ctrs.string_of_ctrs 0 result3);
		      ()
		    end
		  else
		    ();
		  if is_co3 ()
		  then 
		    if !result_U_sound 
                        &&
			begin
			  debug "U for 3DCTRSs is sound for the above CTRS.\n";
			  Cr.is_CR result3
			end
		    then
		      begin
			debug "U(R) is confluent.\n";
			append_output "YES\n";
			print_string "YES\n";
			print_output ()
		      end
                    else
                      if not !competition
                          &&
                         !result_SR_sound 
			  &&
		         begin
			   debug "SR for 3DCTRSs is sound for the above CTRS.\n";
			   Cr.is_CR result2
		         end
		      then
		        begin
			  debug "SR(R) is confluent.\n";
			  append_output "YES\n";
			  print_string "YES\n";
			  print_output ()
		        end
		      else 
		        begin
			  append_output "MAYBE\n";
			  print_string "MAYBE\n";
			  print_output ()
		        end
		  else
		    print_output ()
	    end
      end
    else 
      if is_co3 ()
      then
	begin
	  print_string "usage:\n";
	  print_string "       co3 <filename>\n";
	  print_string "       co3coco <filename>\n";
	  print_string "       co3v <filename>\n";
	  ()
	end
      else
	begin
	  print_string "usage:\n";
	  print_string "       co2 <filename>\n";
	  print_string "       co2un <filename>\n";
	  print_string "       co2u <filename>\n";
	  print_string "       co2srn <filename>\n";
	  print_string "       co2sr <filename>\n";
	  ()
	end
  with
    Finish -> ()
  | Others -> ()
(*
	let result = 
	  try
	    Readtrs.readtrs (Sys.argv.(1))
	  with
	    _ -> raise Unsupported
        in
        Abh.test result
*)
  | e ->
      if !competition
      then  (* for CoCo 2016 *)
	print_string "MAYBE\n"
      else
	begin
	  match e with
	  | Not_3DCTRS ->
	      begin
		debug "UNSUPPORTED\nThe input is not a 3DCTRS.\n";
		print_string "UNSUPPORTED\n"
	      end
	  | Unsupported -> print_string "UNSUPPORTED\n"
	  | Contradiction ->
	      print_string "ERROR: conflicts on CONDITIONTYPE\n"
	  | Convert_NormalOneCTRS.UnexpectedError s
	  | Convert_UWLLDCTRS.UnexpectedError s -> 
	      print_string ("ERROR: unexpected ("^s^")\n")
	  | Failure msg -> 
	      print_string ("ERROR with Failure: "^msg^"\n")
	  | _ -> 
	      print_string "ERROR: unknown\n"
	end;
      print_output ();;
	
