(**
   This is a module for dealing with critical pairs of rewrite systems.
   @author Naoki Nishida
   @since 2012/01/18
   @since 2016/11/22
*)

open Debug;;  
open Util;;  
open Ctrs;;
open Verbose;;
  
(**
   This is an object for debug.
*)
(*
let debug = new debug_print;;
*)
    
(**
   This is an auxiliary function to comput critical pairs.
*)
let rec cp_component l1  (l2,r2) =
  begin
    match l1 with
      Node(FSym _ as f,ls) ->
	let res = 
	  List.map 
	    (fun (ts,sigma) -> (Node(f,ts),sigma)) 
	    (cp_componentlist ls (l2,r2))
	in
	begin
	  try
	    let theta = unify l1 l2 in
	    (r2,theta)::res
	  with
	    Not_found -> res
	end
    | _ -> []
  end
and cp_componentlist ls (l2,r2) =
  begin
    match ls with
      [] -> []
    | t::ts ->
	(List.map 
	   (fun (u,theta) -> (u::ts,theta)) 
	   (cp_component t (l2,r2)))
	@ (List.map 
	     (fun (us,theta) -> (t::us,theta))
	     (cp_componentlist ts (l2,r2)))
  end
and cp_assemble r cps =
  List.map
    (fun (t,theta) ->
      (apply_subst_to_term theta t, apply_subst_to_term theta r))
    cps;;
    
(**
   This computes critical pairs of two rules.
*)
let cpairs_of_rules rl1 rl2 =
  let xs1 = vars_in_rule rl1 in
  Variable.init_gen_var ("x",xs1,[]);
  let rl3 = Rename.rename_rule rl2 in
(*
  print_string ("(COMMENT renamed rule: "^(string_of_rule None rl3)^")\n");
  print_string ("(COMMENT    base rule: "^(string_of_rule None rl1)^")\n");
*)
  let (l3,r3) = (lhs rl1, rhs rl1) in
  let (l4,r4) = (lhs rl3, rhs rl3) in
  cp_assemble r3 (cp_component l3 (l4,r4))
  @ cp_assemble r4 (cp_component l4 (l3,r3));;

(**
    This computes critical pairs of a rule itself.
*)
let selfcpair rl1 =
  let xs1 = vars_in_rule rl1 in
  Variable.init_gen_var ("x",xs1,[]);
  let rl3 = Rename.rename_rule rl1 in
  let (l3,r3) = (lhs rl1, rhs rl1) in
  let (l4,r4) = (lhs rl3, rhs rl3) in
  begin
    match l3 with
      Node(Var _,_) -> raise (Failure "Error in cpair: unexpected.\n")
    | Node(f,ls) ->
	cp_assemble r3 (List.map
			  (fun (ts,sigma) -> (Node(f,ts),sigma))
			  (cp_componentlist ls (l4,r4)))
  end;;

(**
   This computes all critical pairs of a TRS.
*)
let all_cpairs spec = 
  let rec get_cpairs =
    function
	[] -> []
      |	(rl::rs) ->
	  (selfcpair rl) 
	  @ (List.concat (List.map (cpairs_of_rules rl) rs))
	  @ (get_cpairs rs)
  in
  get_cpairs spec.rules;;

(**
   This decides whether arule is overlapping itself.
*)
let is_self_overlapping_rule =
  function
      Rew(l,r,_) -> 
	begin
	  match selfcpair (Rew(l,r,[])) with
	    [] -> false
	  | _ -> true
	end
    | _ -> raise Unsupported;;
	  
(**
   This decides whether two rules are overlapping or not.
   This returns [false] if the two rules are variant.
*)
let is_overlapping_rule rl1 rl2 =
  let xs1 = vars_in_rule rl1 in
  Variable.init_gen_var ("x",xs1,[]);
  let rl3 = Rename.rename_rule rl2 in
  let l1 = lhs rl1 in
  let l2 = lhs rl3 in
  let rec is_ol t =
    function
	Node(Var _,_) -> false
      |	Node(f,us) as u -> 
	  if is_unifiable t u 
	  then true
	  else
	    begin
	      try
		let _ = List.find (is_ol t) us in
		true
	      with
		Not_found -> false
	    end
  in
  let is_inner_ol t =
    function
	Node(Var _,_) -> false
      |	Node(f,us) ->
	  begin
	    try
	      let _ = List.find (is_ol t) us in
	      true
	    with
	      Not_found -> false
	  end
  in
(*
  debug "(COMMENT\n";
  debug  
    ("check whether "^(string_of_term l1)^" and "^(string_of_term l2)^" are unifiable or not.\n");
  debug ")\n";
*)
  if is_variant rl1 rl3 then is_inner_ol l1 l2
  else is_unifiable l1 l2 || is_inner_ol l1 l2 || is_inner_ol l2 l1;;

(**
   This decides whether the rule set is overlapping or not.
*)
let rec is_overlapping =
  function
      [] -> false
    | (rl::rs) ->
	if is_self_overlapping_rule rl
	then true
	else
	  begin
	    try
	      let _ = List.find (is_overlapping_rule rl) rs in
	      true
	    with
	      Not_found -> is_overlapping rs
	  end;;

(**
   This decides whether a rule is overlapping with others.
*)
let is_overlapping_in_rules rl rs =
  begin
    try
      let rl1 = List.find (is_overlapping_rule rl) rs in
(*
      debug "(COMMENT \n";
      debug ((string_of_rule None rl)^" is at least overlapping with "^(string_of_rule None rl1)^".\n");
      debug ")\n";
*)
      true
    with
      Not_found -> 
(*
	debug "(COMMENT \n";
	debug  
	  ((string_of_rule None rl)^" is not overlapping with any other rules.\n");
	debug ")\n";
*)	false
  end;;

(**
   This decides whether a critical pair is joinable.
   Use this function after proving termination of [spec].
*)
let is_joinable spec (s,t) =
(*
  let s_nfs = Reduce.all_nfs spec.rules s in
  let t_nfs = Reduce.all_nfs spec.rules t in
  begin
    match intersection s_nfs t_nfs with
      [] -> false
    | _ -> true
  end;;
*)
  let s_nfs = Reduce.linf spec.rules s in
  let t_nfs = Reduce.linf spec.rules t in
  if s_nfs = t_nfs
  then
    true
  else
    begin
      debug ("Failed to prove joinability of \""^(string_of_term s)^"\" and \""^(string_of_term t)^"\".\n");
      false
    end;;
