(**
   This is a module to provide a function checking whether a condition is infeasible or not.
   
   @since 2015/07/17
   @author Naoki Nishida
*)
 
open Util;; 
open Ctrs;;

(**
   This is a data type for sets of reachable symbols at root positions.
*)
type reachable_syms = Any | Part of fid list;;

(**
   This decides whether a given symbol is reachable to another given symbol at root position.
*)
let reachable_at_root rs s t =
  match (s,t) with
    (Node(FSym f,_),Node(FSym g,_)) -> 
      begin
	let fs = ref [f] in
	let prev = ref !fs in
	let rec reachable rs1 =
	  begin
	    match rs1 with
	      [] -> 
		if !prev = !fs
		then
		  false
		else
		  begin
		    prev := !fs;
		    reachable rs
		  end
	    | (Rew (Node (FSym f1,_),Node(f2,_),c))::rs2
	    | (RelRew (Node (FSym f1,_),Node(f2,_),c))::rs2 ->
		if List.mem f1 !prev 
		then
		  begin
		    match f2 with
		      FSym f3 ->
			if g = f3 
			then true 
			else 
			  begin
			    if List.mem f3 !fs 
			    then
			      ()
			    else
			      fs := f3 :: (!fs);
			    reachable rs2
			  end
		    |	Var _ -> true
		  end
		else
		  reachable rs2
	    |	_ -> raise (Failure "unexpected")
	  end
	in
	f = g || reachable rs
      end
  | _ -> true;;

(**
   This is the CAP known in the DP method.
*)
let cap ds xs s =
  let xs = union xs (vars_in_term s) in
  begin
    Variable.init_gen_var ("x",xs,[]);
    let rec cap0 t =
      begin
	match t with
	  Node(Var _,_) -> t
	| Node(FSym f,ts) ->
	    if List.mem f ds
	    then Node(Var (Variable.gen_var#generate),[])
	    else Node(FSym f,List.map cap0 ts)
      end
    in
    cap0 s
  end;;
  
(**
   This is the REN known in the DP method.
*)
let ren xs s =
  let xs = union xs (vars_in_term s) in
  begin
    Variable.init_gen_var ("x",xs,[]);
    let rec ren0 t =
      begin
	match t with
	  Node(Var _,_) -> Node(Var (Variable.gen_var#generate),[])
	| Node(FSym f,ts) ->
	    Node(FSym f,List.map ren0 ts)
      end
    in
    ren0 s
  end;;

(**
    This is the REN(CAP()) known in the DP method.
*)
let rencap ds xs s = ren xs (cap ds xs s);;
  
(**
  This decides whether evaluation of f-rules is always infinite.
*)
let infinite_eval_form ds rl =
  begin
    match rl with
      Rew (Node(FSym f,_),r,c) 
    | RelRew (Node(FSym f,_),r,c) ->
	if List.length c <= 0
	then false
	else
	  let rec call_f c1 =
	    begin
	      match c1 with
		[] -> false
	      |	(Arrow (Node(FSym g,us),t))::c2 -> 
		  if g = f 
		  then
		    let xs = vars_in_term t in
		    (not (is_unifiable (ren xs (Node(FSym g,List.map (cap ds xs) us))) t))
		  ||
		    call_f c2
		  else
		    call_f c2
	      |	(Arrow _)::c2 ->
		  call_f c2
	      |	_ -> raise (Failure "unexpected")
	    end
	  in
	  call_f c
    | _ -> raise (Failure "unexpected")
  end;;
    
(**
   This decomposes rules to f-rules and others.
*)
let rec decompose_frules f rs =
  match rs with
    [] -> ([],[])
  | rl::rs1 ->
      begin
	match rl with
	  Rew(Node(FSym g,_),_,_)
	| RelRew(Node(FSym g,_),_,_) ->
	    let (rs2,rs3) = decompose_frules f rs1 in
	    if f = g
	    then
	      (rl::rs2,rs3)
	    else
	      (rs2,rl::rs3)
	| _ -> raise (Failure "unexpected")
      end;;
   
(**
   This removes the f-rules if all of them calls f infinitely many times.
*)
let rec rm_infinite_eval_frules ds ds1 rs =
  match (ds1,rs) with
    ([],_) -> rs
  | (f::ds2,_) ->
      let (rs1,rs2) = decompose_frules f rs in
      let rs3 = rm_infinite_eval_frules ds ds2 rs2 in
      if List.for_all (infinite_eval_form ds) rs1
      then
	rs3
      else
	rs1 @ rs3;;

(**
   This decides whether a given condition is infeasible.
   Note that this works Arrow conditions only.
*)
let rec is_infeasible_condition ds rs c = 
  begin
    match c with
      Arrow (s,t) ->
	(not (is_unifiable (rencap ds (vars_in_term t) s) t))
      || 
	(not (reachable_at_root rs s t))
    | _ -> raise (Failure "unexpected")
  end
and is_not_infeasible_condition ds rs c =
  not (is_infeasible_condition ds rs c);;
    
(**
   This decides whether a given condition list is infeasible.
*)
let rec is_infeasible_conditions ds rs c =
  List.exists (is_infeasible_condition ds rs) c
and is_not_infeasible_conditions ds rs c =
  not (is_infeasible_condition ds rs c);;
    
(**
   This decides whether a rule is infeasible.
*)
let rec is_infeasible_rule ds rs rl = 
  match rl with
    Rew (_,_,c)
  | RelRew (_,_,c) ->
      is_infeasible_conditions ds rs c
and is_not_infeasible_rule ds rs rl =
  not (is_infeasible_rule ds rs rl);;
    
(**
   This removes from rules infeasible ones.
*)
let rm_infeasible_rules spec : ctrs =
  let ds = get_dsym spec.rules in
  let rs = List.filter (is_not_infeasible_rule ds spec.rules) spec.rules in
  let ds1 = get_dsym rs in
  let rs1 = rm_infinite_eval_frules ds1 ds1 rs in
  { var = spec.var;
    theories = spec.theories ;
    rules = rs1 ;
    strategy = spec.strategy ; 
    others = spec.others ;
    ctype = spec.ctype ;
  } ;;
    
(**
   This removes from rules infeasible ones as much as possible.
*)
let rm_infeasible_rules_amap spec : ctrs =
  let rec rm rs =
    let ds = get_dsym rs in
    let rs1 = List.filter (is_not_infeasible_rule ds rs) rs in
    let ds1 = get_dsym rs1 in
    let rs2 = rm_infinite_eval_frules ds1 ds1 rs1 in
    if List.length rs = List.length rs2
    then
      rs
    else 
      rm rs2
  in
  let rs = rm spec.rules in
  { var = spec.var;
    theories = spec.theories ;
    rules = rs ;
    strategy = spec.strategy ; 
    others = spec.others ;
    ctype = spec.ctype ;
  } ;;
    
  
