(**
   This is a module to read [.trs] files via [Readtrs_*] modules.
   @author Naoki Nishida
   @since 2010/12/22
*)

open Variable;;
open Ctrs;;

(**
   This converts variable list to [vid list].
*)
let rec read_vardecl vs =
  function
      [] -> vs
    | x::xs -> 
	if List.mem x vs then read_vardecl vs xs else read_vardecl (vs@[x]) xs;;

(**
   This converts [Readtrs_ast.term] to [Ctrs.term].
*)
let rec toTerm xs (Readtrs_ast.Term((_,f),ts)) =
  if List.mem f xs 
  then 
    begin
      match ts with
	[] -> Node(Var (vid_of_string f),[])
      |	_ -> raise (Failure ("Error: variable "^f^" has arguments.\n"))
    end
  else
    Node(FSym (fid_of_string f), List.map (toTerm xs) ts);;

(**
   This converts [Readtrs_ast.theory] to [Ctrs.theory].
*)
let toTheory xs =
  function
      Readtrs_ast.Builtin (s,ss) -> Builtin (snd s,List.map (fun x -> fid_of_string (snd x)) ss)
    | Readtrs_ast.Equations es -> 
	Equations (List.map (fun (s,t) -> (toTerm xs s,toTerm xs t)) es);;

(**
   This converts [Readtrs_ast.cond] to [Ctrs.condition].
*)
let toCondition xs =
  function
      Readtrs_ast.Arrow (s,t) -> Arrow (toTerm xs s, toTerm xs t)
    | Readtrs_ast.Darrow (s,t) -> Darrow (toTerm xs s, toTerm xs t)
    | Readtrs_ast.Equal (s,t) -> Equal (toTerm xs s, toTerm xs t);;

(**
   This converts [Readtrs_ast.rule] to [Ctrs.rule].
*)
let toRule xs =
  function
      Readtrs_ast.Rew (c,l,r) ->
	Rew (toTerm xs l, toTerm xs r, List.map (toCondition xs) c)
    | Readtrs_ast.RelRew (c,l,r) ->
	RelRew (toTerm xs l, toTerm xs r, List.map (toCondition xs) c);;      

(*
   This converts [Readtrs_ast.decl] to [Ctrs.ctrs].
*)
let rec read_decl0 vs ths rs st os ct =
  function
      [] -> 
	let xs = List.map vid_of_string vs in
	{ var = xs;
	  theories = List.map (toTheory vs) ths; 
	  rules = List.map (toRule vs) rs; 
	  strategy = st; 
	  others = os;
	  ctype = ct;
	}
    | (Readtrs_ast.VarDecl xs)::ds ->
	let vs1 = read_vardecl vs (List.map (fun x -> snd x) xs) in
	read_decl0 vs1 ths rs st os ct ds
    | (Readtrs_ast.TheoryDecl ths1)::ds ->
	read_decl0 vs (ths@ths1) rs st os ct ds
    | (Readtrs_ast.RulesDecl rs1)::ds ->
	read_decl0 vs ths (rs@rs1) st os ct ds
    | (Readtrs_ast.StrategyDecl st1)::ds ->
	let st2 =
	  begin
	    match st with
	      None ->
		begin
		  match st1 with
		    Readtrs_ast.Innermost -> Some Innermost
		  | Readtrs_ast.Outermost -> Some Outermost
		  | Readtrs_ast.ContextSensitive css -> 
		      let css1 = 
			List.map (fun (s,ns) -> (fid_of_string (snd s),List.map snd ns)) css
		      in
		      Some (ContextSensitive css1)
		end
	    | _ ->
		raise (Failure "Error: two or more strategies are specified.\n")
	  end
	in
	read_decl0 vs ths rs st2 os ct ds
    | (Readtrs_ast.ConditionTypeDecl c)::ds ->
	let c1 = 
	  begin
	    match c with
	      Readtrs_ast.Semiequational -> Semiequational
	    | Readtrs_ast.Join -> Join
	    | Readtrs_ast.Oriented -> Oriented
	  end
	in
	read_decl0 vs ths rs st os (Some c1) ds
    | (Readtrs_ast.OtherDecl (s,ss))::ds ->
	read_decl0 vs ths rs st (os@[(snd s,ss)]) ct ds;;

(**
   This converts [Readtrs_ast.decl] to [Ctrs.ctrs].
*)
let read_decl ds = read_decl0 [] [] [] None [] None ds;;

(**
   This reflects "CONDITIONTYPE" to rules.
*)
let revisit_ctype_for_cond ct c =
   begin
   match ct with
   None -> c
   | Some Semiequational -> raise Unsupported
   | Some Join ->
   begin
   match c with
   Arrow _ -> raise Contradiction
   | Darrow _ -> c
   | Equal (s,t) -> Darrow (s,t)
   end
   | Some Oriented ->
   begin
   match c with
   Arrow _ -> c
   | Darrow _ -> raise Contradiction
   | Equal (s,t) -> Arrow (s,t)
   end
   end;;

(**
   This reflects "CONDITIONTYPE" to rules.
*)
let revisit_ctype_for_rule ct =
   function
   (Rew (l,r,c)) -> 
   Rew (l,r,List.map (revisit_ctype_for_cond ct) c)
   | (RelRew (l,r,c)) ->
   RelRew (l,r,List.map (revisit_ctype_for_cond ct) c);;

(**
   This reflects "CONDITIONTYPE" to the given CTRS.
*)
let revisit_ctype spec =
   { var = spec.var;
   theories = spec.theories;
   rules = List.map (revisit_ctype_for_rule spec.ctype) spec.rules;
   strategy = spec.strategy;
   others = spec.others;
   ctype = spec.ctype;
   };;

(**
   This reads a [.trs].file and returns [Ctrs.ctrs] data.
*)
let readtrs file =
  let ds = Readtrs_check.check_trs file in
  revisit_ctype (read_decl ds);;

(**
   This reads a string based on a [.trs].file and returns [Ctrs.ctrs] data.
*)
let readtrs_from_string s =
  let ds = Readtrs_check.check_trs_from_string s in
  revisit_ctype (read_decl ds);;
