(**
   This originates with [trs_check.ml].
   This provides a file checker for .trs format.
*)

open Format
open Lexing
open Readtrs_ast

let verbose = ref false

(**
This returns the extension of given file names,
e.g., [extension "test.a.txt" = "txt"].
*)
let extension f =
  try
    let i = String.rindex f '.' in
    String.sub f (i+1) (String.length f - i - 1)
  with Not_found -> ""

(**
This removes the extension from given file names,
e.g., [rm_extension "test.a.txt" = "test.a"].
*)
let rm_extension f =
  try
    let i = String.rindex f '.' in
    String.sub f 0 i 
  with Not_found -> f

(* db output *)

let db_output = ref false

let db_ch = ref stdout

let db_fmt = ref std_formatter

let db_trs_pbs = ref []

let make_id dir file =
  let s = dir ^ "." ^ file in
  for i=0 to String.length s - 1 do
    match String.get s i with
      | '0'..'9' | 'A' .. 'Z' | 'a' .. 'z' | '_' | '.' | '-' -> ()
      | '/' -> String.set s i '.'
      | _ -> String.set s i '_'
  done;
  s

let db_pb pbs f =
  let dir = Filename.dirname f in
  let file = Filename.basename f in
  let name = rm_extension file in
  let id = make_id dir file in
  fprintf !db_fmt "%s = {@." id;
  pbs := id :: !pbs;
  fprintf !db_fmt "name  = \"%s - %s\";@." dir name;
  fprintf !db_fmt "file  = \"%s\";@." f

let get_categories ast =
  List.fold_left
    (fun (cond,t,i,cs,rel) decl ->
       match decl with
	 | RulesDecl l -> 
	     let cond =
	       List.fold_left
		 (fun cond r -> 
		    match r with
		      | Rew([],_,_) -> cond
		      | Rew(_,_,_) -> true
		      | RelRew(_,_,_) -> cond)
		 cond l
	     in 
	     let rel =
	       List.fold_left
		 (fun rel r -> 
		    match r with
		      | Rew(_,_,_) -> rel
		      | RelRew(_,_,_) -> true)
		 rel l
	     in 
	     (cond,t,i,cs,rel)
	 | TheoryDecl _ -> (cond,true,i,cs,rel)
	 | StrategyDecl Innermost -> (cond,t,true,cs,rel)
	 | StrategyDecl Outermost -> (cond,t,i,cs,rel)
	 | StrategyDecl (ContextSensitive _) -> (cond,t,i,true,rel)
	 |_ -> (cond,t,i,cs,rel))
    (false,false,false,false,false)
    ast

let check_has f b1 b2 msg =
  if b1 && b2 then
    printf "Warning: problem %s has both %s.@." f msg 

let db_trs f ast = 
  db_pb db_trs_pbs f;
  let (has_cond,has_theory,has_innermost,has_context_sensitive,has_rel) =
    get_categories ast
  in
  check_has f has_cond has_theory "conditions and theory";
  check_has f has_cond has_innermost "conditions and innermost strategy";
  check_has f has_cond has_context_sensitive "conditions and CS strategy";
  check_has f has_theory has_innermost "theory and innermost strategy";
  check_has f has_theory has_context_sensitive "theory and CS strategy";
  check_has f has_innermost has_context_sensitive "innermost and CS strategy";
  fprintf !db_fmt "conditional = %b;@." has_cond;
  fprintf !db_fmt "theory  = %b;@." has_theory;
  fprintf !db_fmt "innermost  = %b;@." has_innermost;
  fprintf !db_fmt "contextsensitive = %b;@." has_context_sensitive;
  fprintf !db_fmt "relative = %b;@." has_rel;
  fprintf !db_fmt "}@.@."

let db_list name l =
  match l with
    | [] -> fprintf !db_fmt "%s = [ ];@." name;
    | p::r ->
	  fprintf !db_fmt "@[%s = [ %s" name p;
	  List.iter (fun f -> fprintf !db_fmt ";@ %s" f) r;
	  fprintf !db_fmt "];@]@."

let db_main () =
  fprintf !db_fmt "_main = {@.";
  fprintf !db_fmt "trs_tools = [ aprove ; cariboo ; cime ; matchbox ; ttt] ;@.";
  db_list "trs_pbs" !db_trs_pbs;
  fprintf !db_fmt "}@.";
  close_out !db_ch;
  printf "Database written@.";
  printf "Number of TRS pbs: %d@." (List.length !db_trs_pbs)


(* main *)

let trs_files = ref []

let rec parse_args args =
  match args with
    | [] -> ()
    | "-v"::rem -> verbose := true; parse_args rem
    | "-db"::f::rem -> 
	db_output := true; 
	db_ch := open_out f; 	
	db_fmt := formatter_of_out_channel !db_ch;
	parse_args rem
    | "-trs"::f::rem -> 
	trs_files := f :: !trs_files; parse_args rem
    | f::rem ->
	begin
	  match extension f with
	    | "trs" ->
		trs_files := f :: !trs_files; parse_args rem
	    | _ ->
		eprintf "don't known what to do with argument %s@." f;
		exit 2;
	end
	
let read_trs f lb =
  lb.lex_curr_p <-
    { pos_fname = f ; pos_lnum = 1; pos_cnum = 0; pos_bol = 0};
  let ast = Readtrs_parser.spec Readtrs_lexer.token lb in
  if !verbose then printf "parse ok@.";
  ast
      
let check_trs0 f ast =
  try
    let env = Readtrs_sem.spec ast in
    if !verbose then 
      begin
	printf "semantics ok, function symbols are:@.";
	List.iter
	  (fun (id,arity) ->
	     match arity with
	       | Readtrs_sem.Var -> ()
	       | Readtrs_sem.Function x -> printf "%s: %d@." id !x)
	  env
      end;
    if !db_output then db_trs f ast;
    ast (* added *)
  with 
    | Readtrs_sem.Sem_error(loc,msg) ->
	eprintf "%a: %s.@." Readtrs_ast.print_loc loc msg;
	exit 1

let check_trs f =
  let c = open_in f in
  let lb = from_channel c in
  try
    let ast = read_trs f lb in
    close_in c;
    let env = check_trs0 f ast in
    env
  with 
    | Parsing.Parse_error ->
	let loc = Lexing.lexeme_start_p lb in
	eprintf "%a: syntax error.@." Readtrs_ast.print_loc loc;
	close_in c;
	exit 1
    | Readtrs_lexer.Lexing_error msg ->
	let loc = Lexing.lexeme_start_p lb in
	eprintf "%a: lexer error, %s.@." Readtrs_ast.print_loc loc msg;
	close_in c;
	exit 1

let check_trs_from_string s =
  let lb = from_string s in
  let f = "input_string" in
  try
    let ast = read_trs f lb in
    let env = check_trs0 f ast in
    env
  with 
    | Parsing.Parse_error ->
	let loc = Lexing.lexeme_start_p lb in
	eprintf "%a: syntax error.@." Readtrs_ast.print_loc loc;
	exit 1
    | Readtrs_lexer.Lexing_error msg ->
	let loc = Lexing.lexeme_start_p lb in
	eprintf "%a: lexer error, %s.@." Readtrs_ast.print_loc loc msg;
	exit 1

let check_trs_from_stdin f =
  let lb = from_channel stdin in
  let f = "stdin" in
  try
    let ast = read_trs f lb in
    let env = check_trs0 f ast in
    env
  with 
    | Parsing.Parse_error ->
	let loc = Lexing.lexeme_start_p lb in
	eprintf "%a: syntax error.@." Readtrs_ast.print_loc loc;
	exit 1
    | Readtrs_lexer.Lexing_error msg ->
	let loc = Lexing.lexeme_start_p lb in
	eprintf "%a: lexer error, %s.@." Readtrs_ast.print_loc loc msg;
	exit 1

(*
let main () =
  try
    parse_args (List.tl (Array.to_list Sys.argv));
    List.iter check_trs !trs_files;
    if !db_output then db_main ();
    exit 0
  with
  | Sys_error msg ->
      eprintf "System error: %s.@." msg;
      exit 1
  | Failure msg ->
      eprintf "Failure: %s.@." msg;
      exit 1
	
let _ = Printexc.catch main ()
	     
*)
